"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = require("ramda");
const Array = require("fp-ts/lib/Array");
const Either = require("fp-ts/lib/Either");
const Option = require("fp-ts/lib/Option");
const Task = require("fp-ts/lib/Task");
const TaskEither = require("fp-ts/lib/TaskEither");
const function_1 = require("fp-ts/lib/function");
const pipeable_1 = require("fp-ts/lib/pipeable");
const ast = require("./ast");
const fp_utils_1 = require("./fp-utils");
const const_utils_1 = require("./const-utils");
const field_nullability_1 = require("./field-nullability");
const parser_1 = require("./parser");
const types_1 = require("./types");
const Warn = require("./Warn");
const InferM = require("./InferM");
function virtualField(name, nullability) {
    return { name, nullability };
}
function inferStatementNullability(client, statement) {
    return pipeable_1.pipe(TaskEither.fromEither(parser_1.parse(statement.sql)), TaskEither.chain(astNode => pipeable_1.pipe(getParamNullability(client, astNode), InferM.fromTaskEither, InferM.chain(paramNullability => pipeable_1.pipe(inferColumnNullability(client, statement, paramNullability, astNode), InferM.map(stmt => applyParamNullability(stmt, paramNullability)))), InferM.map(stmt => inferRowCount(stmt, astNode)))), TaskEither.getOrElse(parseErrorStr => Task.of(Warn.warning(statement, 'The internal SQL parser failed to parse the SQL statement.', `Parse error: ${parseErrorStr}`))));
}
exports.inferStatementNullability = inferStatementNullability;
function inferColumnNullability(client, statement, paramNullability, tree) {
    return pipeable_1.pipe(getOutputColumns(client, [], paramNullability, tree), InferM.chain(outputColumns => InferM.fromEither(applyColumnNullability(statement, outputColumns))));
}
exports.inferColumnNullability = inferColumnNullability;
function getOutputColumns(client, outsideCTEs, paramNullability, tree) {
    return ast.walk(tree, {
        select: ({ ctes, body, setOps }) => pipeable_1.pipe(combineVirtualTables(outsideCTEs, getVirtualTablesForWithQueries(client, paramNullability, ctes)), InferM.chain(combinedCTEs => inferSetOpsOutput(client, combinedCTEs, paramNullability, body, setOps))),
        insert: ({ table, as, returning }) => pipeable_1.pipe(getSourceColumnsForTable(client, outsideCTEs, table, as), InferM.chain(sourceColumns => inferSelectListOutput(client, outsideCTEs, sourceColumns, paramNullability, [], returning))),
        update: ({ ctes, table, as, from, where, returning }) => pipeable_1.pipe(combineVirtualTables(outsideCTEs, getVirtualTablesForWithQueries(client, paramNullability, ctes)), InferM.chain(combinedCTEs => combineSourceColumns(getSourceColumnsForTableExpr(client, combinedCTEs, paramNullability, from), getSourceColumnsForTable(client, combinedCTEs, table, as))), InferM.chain(sourceColumns => inferSelectListOutput(client, outsideCTEs, sourceColumns, paramNullability, [where], returning))),
        delete: ({ table, as, where, returning }) => pipeable_1.pipe(getSourceColumnsForTable(client, outsideCTEs, table, as), InferM.chain(sourceColumns => inferSelectListOutput(client, outsideCTEs, sourceColumns, paramNullability, [where], returning))),
    });
}
function applyColumnNullability(stmt, outputColumns) {
    if (outputColumns.length != stmt.columns.length) {
        return Either.left(`BUG: Non-equal number of columns: \
inferred ${outputColumns.length}, actual ${stmt.columns.length}`);
    }
    const inferredColumnNames = outputColumns.map(({ name }) => name).join(', ');
    const actualColumnNames = stmt.columns.map(({ name }) => name).join(', ');
    if (inferredColumnNames != actualColumnNames) {
        return Either.left(`BUG: Inferred output column names do not equal \
actual output column names: inferred "${inferredColumnNames}", \
actual: "${actualColumnNames}"`);
    }
    return Either.right(Object.assign(Object.assign({}, stmt), { columns: R.zipWith((column, inferred) => {
            switch (inferred.nullability.kind) {
                case 'Any':
                    return Object.assign(Object.assign({}, column), { nullable: inferred.nullability.nullable });
                case 'Array':
                    return Object.assign(Object.assign({}, column), { type: types_1.ValueType.array(column.type.oid, inferred.nullability.elemNullable), nullable: inferred.nullability.nullable });
            }
        }, stmt.columns, outputColumns) }));
}
function inferSetOpsOutput(client, outsideCTEs, paramNullability, first, setOps) {
    // fp-ts's foldM is not stack safe so a manual loop is needed
    return () => __awaiter(this, void 0, void 0, function* () {
        let curr = yield inferSelectBodyOutput(client, outsideCTEs, paramNullability, first)();
        if (Either.isLeft(curr)) {
            return curr;
        }
        let result = curr.right;
        for (const setOp of setOps) {
            const next = yield inferSelectBodyOutput(client, outsideCTEs, paramNullability, setOp.select)();
            if (Either.isLeft(next)) {
                return next;
            }
            const columns = next.right;
            if (result.payload.length != columns.payload.length) {
                return Either.left(`Inequal number of columns in ${setOp.op}`);
            }
            // EXCEPT has no effect on nullability of the output, because
            // its output is not indluded
            if (setOp.op !== 'EXCEPT') {
                result = pipeable_1.pipe(Warn.of((aa) => (bb) => pipeable_1.pipe(Array.zip(aa, bb), Array.map(([a, b]) => ({
                    // Column names are determined by the first SELECT
                    name: a.name,
                    nullability: field_nullability_1.FieldNullability.disjunction(a.nullability)(b.nullability),
                })))), Warn.ap(result), Warn.ap(columns));
            }
        }
        return Either.right(result);
    });
}
function inferSelectBodyOutput(client, outsideCTEs, paramNullability, body) {
    return pipeable_1.pipe(getSourceColumnsForTableExpr(client, outsideCTEs, paramNullability, body.from), InferM.chain(sourceColumns => inferSelectListOutput(client, outsideCTEs, sourceColumns, paramNullability, [body.where, body.having], body.selectList)));
}
function inferSelectListOutput(client, outsideCTEs, sourceColumns, paramNullability, conditions, selectList) {
    return pipeable_1.pipe(InferM.right(pipeable_1.pipe(conditions.map(cond => getNonNullSubExpressionsFromRowCond(cond)), Array.flatten)), InferM.chain(nonNullExpressions => pipeable_1.pipe(fp_utils_1.traverseAIM(selectList, item => inferSelectListItemOutput(client, outsideCTEs, sourceColumns, paramNullability, nonNullExpressions, item)), InferM.map(R.flatten))));
}
function inferSelectListItemOutput(client, outsideCTEs, sourceColumns, paramNullability, nonNullExpressions, selectListItem) {
    return ast.SelectListItem.walk(selectListItem, {
        allFields: () => InferM.fromEither(pipeable_1.pipe(
        // hidden columns aren't selected by SELECT *
        findNonHiddenSourceColumns(sourceColumns), Either.map(columns => applyExpressionNonNullability(nonNullExpressions, columns)), Either.map(columns => columns.map(column => ({
            name: column.columnName,
            nullability: column.nullability,
        }))))),
        allTableFields: ({ tableName }) => InferM.fromEither(pipeable_1.pipe(findNonHiddenSourceTableColumns(tableName, sourceColumns), Either.map(columns => applyExpressionNonNullability(nonNullExpressions, columns)), Either.map(columns => columns.map(column => ({
            name: column.columnName,
            nullability: column.nullability,
        }))))),
        selectListExpression: ({ expression, as }) => pipeable_1.pipe(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExpressions, expression), InferM.map(exprNullability => [
            virtualField(as || inferExpressionName(expression), exprNullability),
        ])),
    });
}
function isColumnNonNullable(nonNullableColumns, sourceColumn) {
    return nonNullableColumns.some(nonNull => nonNull.tableName
        ? sourceColumn.tableAlias === nonNull.tableName
        : true && sourceColumn.columnName === nonNull.columnName);
}
function applyExpressionNonNullability(nonNullableExpressions, sourceColumns) {
    const nonNullableColumns = pipeable_1.pipe(nonNullableExpressions, R.map(expr => ast.Expression.walkSome(expr, Option.none, {
        columnRef: ({ column }) => Option.some({ tableName: null, columnName: column }),
        tableColumnRef: ({ table, column }) => Option.some({ tableName: table, columnName: column }),
    })), Array.filterMap(function_1.identity));
    return sourceColumns.map(sourceColumn => (Object.assign(Object.assign({}, sourceColumn), { nullability: isColumnNonNullable(nonNullableColumns, sourceColumn)
            ? field_nullability_1.FieldNullability.any(false)
            : sourceColumn.nullability })));
}
function inferExpressionName(expression) {
    return ast.Expression.walkSome(expression, '?column?', {
        columnRef: ({ column }) => column,
        tableColumnRef: ({ column }) => column,
    });
}
function inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, expression) {
    const anyTE = function_1.flow(field_nullability_1.FieldNullability.any, Warn.of, TaskEither.right);
    const arrayTE = function_1.flow(field_nullability_1.FieldNullability.array, Warn.of, TaskEither.right);
    if (nonNullExprs.some(nonNull => ast.Expression.equals(expression, nonNull))) {
        // This expression is guaranteed to be not NULL by a
        // `WHERE EXPR IS NOT NULL` clause
        return anyTE(false);
    }
    return ast.Expression.walk(expression, {
        // A column reference may evaluate to NULL if the column doesn't
        // have a NOT NULL constraint
        tableColumnRef: ({ table, column }) => pipeable_1.pipe(InferM.fromEither(findSourceTableColumn(table, column, sourceColumns)), InferM.map(column => column.nullability)),
        // A column reference may evaluate to NULL if the column doesn't
        // have a NOT NULL constraint
        columnRef: ({ column }) => pipeable_1.pipe(InferM.fromEither(findSourceColumn(column, sourceColumns)), InferM.map(column => column.nullability)),
        // A unary operator has two options:
        //
        // - The operator is known to be NULL safe: it returns NULL only
        //   if its operand is NULL
        //
        // - The operator is not NULL safe: it can return NULL even if its
        // - operand is not NULL
        unaryOp: ({ op, operand }) => {
            switch (const_utils_1.operatorNullSafety(op)) {
                case 'safe':
                    return inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, operand);
                case 'unsafe':
                    return anyTE(true);
                case 'neverNull':
                    return anyTE(false);
                case null:
                    return TaskEither.right(Warn.warning(field_nullability_1.FieldNullability.any(true), `Unknown operator '${op}'`));
            }
        },
        // A binary operator has two options:
        //
        // - The operator is known to be NULL safe: it returns NULL only
        //   if any of its operands is NULL
        //
        // - The function is not NULL safe: it can return NULL even if all
        //   of its operands are non-NULL
        binaryOp: ({ lhs, op, rhs }) => {
            // AND and OR are unsafe from the result side (e.g. FALSE AND
            // NULL => NULL), but if both args are non-null, then the
            // result is also guaranteed to be non-null.
            const nullSafety = op == 'AND' || op == 'OR' ? 'safe' : const_utils_1.operatorNullSafety(op);
            switch (nullSafety) {
                case 'safe':
                    return pipeable_1.pipe(InferM.right(field_nullability_1.FieldNullability.disjunction), InferM.ap(pipeable_1.pipe(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, lhs))), InferM.ap(pipeable_1.pipe(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, rhs))));
                case 'unsafe':
                    return anyTE(true);
                case 'neverNull':
                    return anyTE(false);
                case null:
                    return TaskEither.right(Warn.warning(field_nullability_1.FieldNullability.any(true), `Unknown operator '${op}'`));
            }
        },
        ternaryOp: ({ lhs, op, rhs1, rhs2 }) => {
            switch (const_utils_1.operatorNullSafety(op)) {
                case 'safe':
                    return pipeable_1.pipe(InferM.right((a) => (b) => (c) => field_nullability_1.FieldNullability.disjunction(field_nullability_1.FieldNullability.disjunction(a)(b))(c)), InferM.ap(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, lhs)), InferM.ap(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, rhs1)), InferM.ap(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, rhs2)));
                case 'unsafe':
                    return anyTE(true);
                case 'neverNull':
                    return anyTE(false);
                case null:
                    return TaskEither.right(Warn.warning(field_nullability_1.FieldNullability.any(true), `Unknown operator '${op}'`));
            }
        },
        // EXISTS (subquery) never returns NULL
        existsOp: () => anyTE(false),
        // A function call has two options:
        //
        // - The function is known to be NULL safe: it returns NULL only
        //   if any of its arguments is NULL
        //
        // - The function is not NULL safe: it can return NULL even if all
        //   of its arguments are non-NULL
        //
        functionCall: ({ funcName, argList }) => {
            switch (const_utils_1.functionNullSafety(funcName)) {
                case 'safe':
                    return pipeable_1.pipe(fp_utils_1.traverseATE(argList, arg => inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, arg)), TaskEither.map(argNullability => pipeable_1.pipe(fp_utils_1.sequenceAW(argNullability), Warn.map(an => field_nullability_1.FieldNullability.any(an.some(nullability => nullability.nullable))))));
                case 'unsafe':
                    return anyTE(true);
                case 'neverNull':
                    return anyTE(false);
                case null:
                    return TaskEither.right(Warn.warning(field_nullability_1.FieldNullability.any(true), `Unknown function '${funcName}'`));
            }
        },
        // expr IN (subquery) returns NULL if expr is NULL
        inOp: ({ lhs }) => inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, lhs),
        // ARRAY(subquery) is never null as a whole. The nullability of
        // the inside depends on the inside select list expression
        arraySubQuery: ({ subquery }) => pipeable_1.pipe(getOutputColumns(client, outsideCTEs, paramNullability, subquery), InferM.chain(columns => {
            if (columns.length != 1)
                return TaskEither.left('subquery must return only one column');
            return arrayTE(
            // An array constructed from a subquery is never NULL itself
            false, 
            // Element nullability depends on the subquery column nullability
            columns[0].nullability.nullable);
        })),
        case: ({ branches, else_ }) => {
            if (else_ === null) {
                // No ELSE branch => Rows that match none of the branches
                // will be NULL
                return anyTE(true);
            }
            return pipeable_1.pipe(InferM.right((branchNullabilities) => (elseNullability) => branchNullabilities.reduce((a, b) => field_nullability_1.FieldNullability.disjunction(a)(b), elseNullability)), InferM.ap(pipeable_1.pipe(branches.map(({ condition, result }) => {
                const nonNullExprsByCond = getNonNullSubExpressionsFromRowCond(condition);
                return inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, [...nonNullExprsByCond, ...nonNullExprs], result);
            }), fp_utils_1.sequenceAIM)), InferM.ap(inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, else_)));
        },
        // A type cast evaluates to NULL if the expression to be casted is
        // NULL.
        typeCast: ({ lhs }) => inferExpressionNullability(client, outsideCTEs, sourceColumns, paramNullability, nonNullExprs, lhs),
        // NULL is the only nullable constant
        constant: ({ valueText }) => anyTE(valueText === 'NULL'),
        // By default, a parameter is non-nullable, but param
        // nullability infering may have overridden the default.
        parameter: ({ index }) => {
            const nullability = paramNullability.find(val => val.index === index);
            if (nullability !== undefined)
                return anyTE(nullability.nullable);
            return anyTE(false);
        },
    });
}
function getNonNullSubExpressionsFromRowCond(expression, logicalNegation = false) {
    if (expression == null) {
        return [];
    }
    return ast.Expression.walkSome(expression, [], {
        columnRef: () => {
            return [expression];
        },
        tableColumnRef: () => {
            return [expression];
        },
        unaryOp: ({ op, operand }) => {
            if (op === 'IS NOT NULL' || op === 'NOTNULL') {
                // IS NOT NULL / NOTNULL promise that the operand is not null
                return getNonNullSubExpressionsFromRowCond(operand, logicalNegation);
            }
            if (op === 'NOT') {
                // Track logical negation across NOTs
                return getNonNullSubExpressionsFromRowCond(operand, !logicalNegation);
            }
            if (const_utils_1.operatorNullSafety(op) === 'safe') {
                // For safe operators, the operator must non-nullable for the
                // result to evaluate to non-null
                return getNonNullSubExpressionsFromRowCond(operand, logicalNegation);
            }
            // Otherwise, the whole expression is non-null because it must
            // evaluate to true, but cannot say anything about the operands
            return [expression];
        },
        binaryOp: ({ lhs, op, rhs }) => {
            if (op === 'AND') {
                if (logicalNegation) {
                    // `FALSE AND NULL` evaluates to NULL => NOT (FALSE AND
                    // NULL) evaluates to true, so we cannot say anything about
                    // the right hand side!
                    return [...getNonNullSubExpressionsFromRowCond(lhs, logicalNegation)];
                }
                else {
                    // `a AND b` evaluates to TRUE
                    return [
                        ...getNonNullSubExpressionsFromRowCond(lhs, logicalNegation),
                        ...getNonNullSubExpressionsFromRowCond(rhs, logicalNegation),
                    ];
                }
            }
            if (op === 'AND' || const_utils_1.operatorNullSafety(op) === 'safe') {
                // For safe operators, both sides must be non-nullable for the
                // result to be non-nullable.
                return [
                    ...getNonNullSubExpressionsFromRowCond(lhs, logicalNegation),
                    ...getNonNullSubExpressionsFromRowCond(rhs, logicalNegation),
                ];
            }
            // Otherwise, the whole expression is non-null because it must
            // evaluate to true, but cannot say anything about the operands
            return [expression];
        },
        ternaryOp: ({ lhs, op, rhs1, rhs2 }) => {
            if (const_utils_1.operatorNullSafety(op) === 'safe') {
                // For safe operators, all operands must be non-nullable for the
                // result to be non-nullable.
                return [
                    ...getNonNullSubExpressionsFromRowCond(lhs, logicalNegation),
                    ...getNonNullSubExpressionsFromRowCond(rhs1, logicalNegation),
                    ...getNonNullSubExpressionsFromRowCond(rhs2, logicalNegation),
                ];
            }
            // Otherwise, the whole expression is non-null because it must
            // evaluate to true, but cannot say anything about the operands
            return [expression];
        },
        functionCall: ({ funcName, argList }) => {
            if (const_utils_1.functionNullSafety(funcName) === 'safe') {
                return pipeable_1.pipe(argList, Array.map(arg => getNonNullSubExpressionsFromRowCond(arg, logicalNegation)), Array.flatten);
            }
            return [expression];
        },
    });
}
function getParamNullability(client, tree) {
    return ast.walk(tree, {
        select: () => TaskEither.right([]),
        insert: ({ table, columns, values }) => pipeable_1.pipe(TaskEither.right(combineParamNullability), TaskEither.ap(TaskEither.right(findParamsFromValues(values, columns.length))), TaskEither.ap(findInsertColumns(client, table, columns))),
        update: ({ table, updates }) => findParamNullabilityFromUpdates(client, table, updates),
        delete: () => TaskEither.right([]),
    });
}
function findParamsFromValues(values, numInsertColumns) {
    return ast.Values.walk(values, {
        defaultValues: () => [R.repeat(Option.none, numInsertColumns)],
        exprValues: ({ valuesList }) => valuesList.map(values => values.map(paramIndexFromExpr)),
    });
}
function findParamNullabilityFromUpdates(client, table, updates) {
    return pipeable_1.pipe(client.getTable(table.schema, table.table), TaskEither.chain(dbTable => TaskEither.fromEither(fp_utils_1.traverseAE(updates, update => updateToParamNullability(dbTable, update)))), TaskEither.map(paramNullabilities => pipeable_1.pipe(paramNullabilities, Array.filterMap(function_1.identity))));
}
function paramIndexFromExpr(expression) {
    return pipeable_1.pipe(Option.fromNullable(expression), Option.chain(nonNullExpr => ast.Expression.walkSome(nonNullExpr, Option.none, {
        parameter: paramExpr => Option.some(paramExpr.index - 1),
    })));
}
function updateToParamNullability(dbTable, update) {
    return pipeable_1.pipe(Either.right(makeParamNullability), Either.ap(Either.right(paramIndexFromExpr(update.value))), Either.ap(findTableColumn(update.columnName, dbTable)));
}
const makeParamNullability = (index) => (column) => pipeable_1.pipe(index, Option.map(index => ({ index, nullable: column.nullable })));
function findInsertColumns(client, table, columnNames) {
    return pipeable_1.pipe(client.getTable(table.schema, table.table), TaskEither.chain(dbTable => TaskEither.fromEither(fp_utils_1.traverseAE(columnNames, columnName => findTableColumn(columnName, dbTable)))));
}
const combineParamNullability = (valuesListParams) => (targetColumns) => {
    return pipeable_1.pipe(valuesListParams.map(valuesParams => R.zip(targetColumns, valuesParams).map(([column, param]) => pipeable_1.pipe(param, Option.map(index => ({ index, nullable: column.nullable }))))), R.flatten, Array.filterMap(function_1.identity));
};
function applyParamNullability(stmt, paramNullability) {
    // paramNullability may contain multiple records for each param. If
    // any of the records states that the param is nullable, then it is
    // nullable.
    const nullability = R.range(0, stmt.params.length).map(index => paramNullability
        .filter(record => record.index === index)
        .some(record => record.nullable));
    return Object.assign(Object.assign({}, stmt), { params: R.zipWith((param, nullable) => (Object.assign(Object.assign({}, param), { nullable })), stmt.params, nullability) });
}
function inferRowCount(statement, astNode) {
    const rowCount = ast.walk(astNode, {
        select: ({ body, setOps, limit }) => {
            if (setOps.length === 0 && body.from === null) {
                // No UNION/INTERSECT/EXCEPT, no FROM clause => one row
                return 'one';
            }
            if (limit && limit.count && isConstantExprOf('1', limit.count)) {
                // LIMIT 1 => zero or one row
                return 'zeroOrOne';
            }
            return 'many';
        },
        insert: ({ values, returning }) => ast.Values.walk(values, {
            // INSERT INTO xxx DEFAULT VALUES always creates a single row
            defaultValues: () => 'one',
            exprValues: exprValues => returning.length
                ? // Check the length of the VALUES expression list
                    exprValues.valuesList.length === 1
                        ? 'one'
                        : 'many'
                : // No RETURNING, no output
                    'zero',
        }),
        update: ({ returning }) => returning.length
            ? 'many'
            : // No RETURNING, no output
                'zero',
        delete: ({ returning }) => returning.length
            ? 'many'
            : // No RETURNING, no output
                'zero',
    });
    return Object.assign(Object.assign({}, statement), { rowCount });
}
function getVirtualTablesForWithQueries(client, paramNullability, withQueries) {
    return () => __awaiter(this, void 0, void 0, function* () {
        const virtualTables = [];
        let warnings = [];
        for (const withQuery of withQueries) {
            // "Virtual tables" from previous WITH queries are available
            const virtualTable = pipeable_1.pipe(yield getOutputColumns(client, virtualTables, paramNullability, withQuery.query)(), Either.map(columnsWithWarnings => pipeable_1.pipe(columnsWithWarnings, Warn.map(columns => ({ name: withQuery.as, columns })))));
            if (Either.isLeft(virtualTable)) {
                return virtualTable;
            }
            virtualTables.push(virtualTable.right.payload);
            warnings = warnings.concat(virtualTable.right.warnings);
        }
        return Either.right(Warn.make(virtualTables, warnings));
    });
}
function combineVirtualTables(outsideCTEs, ctes) {
    return pipeable_1.pipe(ctes, InferM.map(virtualTables => [...outsideCTEs, ...virtualTables]));
}
function getSourceColumnsForTable(client, ctes, table, as) {
    if (table.schema == null) {
        // Try to find a matching CTE
        const result = ctes.find(virtualTable => virtualTable.name === table.table);
        if (result)
            return InferM.right(result.columns.map(col => ({
                tableAlias: as || table.table,
                columnName: col.name,
                nullability: col.nullability,
                hidden: false,
            })));
    }
    // No matching CTE, try to find a database table
    return pipeable_1.pipe(client.getTable(table.schema, table.table), TaskEither.map(table => table.columns.map(col => ({
        tableAlias: as || table.name,
        columnName: col.name,
        nullability: field_nullability_1.FieldNullability.any(col.nullable),
        hidden: col.hidden,
    }))), InferM.fromTaskEither);
}
function getSourceColumnsForTableExpr(client, ctes, paramNullability, tableExpr, setNullable = false) {
    if (!tableExpr) {
        return InferM.right([]);
    }
    return pipeable_1.pipe(ast.TableExpression.walk(tableExpr, {
        table: ({ table, as }) => getSourceColumnsForTable(client, ctes, table, as),
        subQuery: ({ query, as }) => getSourceColumnsForSubQuery(client, ctes, paramNullability, query, as),
        crossJoin: ({ left, right }) => combineSourceColumns(getSourceColumnsForTableExpr(client, ctes, paramNullability, left, false), getSourceColumnsForTableExpr(client, ctes, paramNullability, right, false)),
        qualifiedJoin: ({ left, joinType, right }) => combineSourceColumns(getSourceColumnsForTableExpr(client, ctes, paramNullability, left, 
        // RIGHT or FULL JOIN -> The left side columns becomes nullable
        joinType === 'RIGHT' || joinType === 'FULL'), getSourceColumnsForTableExpr(client, ctes, paramNullability, right, 
        // LEFT or FULL JOIN -> The right side columns becomes nullable
        joinType === 'LEFT' || joinType === 'FULL')),
    }), InferM.map(sourceColumns => setNullable ? setSourceColumnsAsNullable(sourceColumns) : sourceColumns));
}
function getSourceColumnsForSubQuery(client, outsideCTEs, paramNullability, subquery, as) {
    return pipeable_1.pipe(getOutputColumns(client, outsideCTEs, paramNullability, subquery), InferM.map(columns => columns.map(column => ({
        tableAlias: as,
        columnName: column.name,
        nullability: column.nullability,
        hidden: false,
    }))));
}
function setSourceColumnsAsNullable(sourceColumns) {
    return sourceColumns.map(col => (Object.assign(Object.assign({}, col), { nullability: Object.assign(Object.assign({}, col.nullability), { nullable: true }) })));
}
function combineSourceColumns(...sourceColumns) {
    return pipeable_1.pipe(sourceColumns, fp_utils_1.sequenceAIM, InferM.map(R.flatten));
}
function isConstantExprOf(expectedValue, expr) {
    return ast.Expression.walkSome(expr, false, {
        constant: ({ valueText }) => valueText === expectedValue,
    });
}
function findNonHiddenSourceColumns(sourceColumns) {
    return pipeable_1.pipe(sourceColumns.filter(col => !col.hidden), Either.fromPredicate(result => result.length > 0, () => `No columns`));
}
function findNonHiddenSourceTableColumns(tableName, sourceColumns) {
    return pipeable_1.pipe(findNonHiddenSourceColumns(sourceColumns), Either.map(sourceColumns => sourceColumns.filter(col => col.tableAlias === tableName)), Either.chain(result => result.length > 0
        ? Either.right(result)
        : Either.left(`No visible columns for table ${tableName}`)));
}
function findSourceTableColumn(tableName, columnName, sourceColumns) {
    return pipeable_1.pipe(sourceColumns.find(source => source.tableAlias === tableName && source.columnName === columnName), Either.fromNullable(`Unknown column ${tableName}.${columnName}`));
}
function findSourceColumn(columnName, sourceColumns) {
    return pipeable_1.pipe(sourceColumns.find(col => col.columnName === columnName), Either.fromNullable(`Unknown column ${columnName}`));
}
function findTableColumn(columnName, table) {
    return pipeable_1.pipe(table.columns.find(column => column.name === columnName), Either.fromNullable(`Unknown column ${columnName}`));
}
