"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typed_parser_1 = require("typed-parser");
const ast_1 = require("../ast");
const token_1 = require("./token");
const utils_1 = require("./utils");
// We need to take the expression parser as an argument to avoid cyclic
// imports
function specialFunctionCall(primaryExpr) {
    function specialFunctionParser(funcName, argsParser) {
        return typed_parser_1.seq((_, argList) => ast_1.Expression.createFunctionCall(funcName, argList), token_1.expectIdentifier(funcName), utils_1.parenthesized(argsParser));
    }
    const overlayFunction = specialFunctionParser('overlay', typed_parser_1.seq((a1, _placing, a2, _from, a3, _for, a4) => [a1, a2, a3, a4], primaryExpr, token_1.expectIdentifier('placing'), primaryExpr, token_1.expectIdentifier('from'), primaryExpr, token_1.expectIdentifier('for'), primaryExpr));
    const positionFunction = specialFunctionParser('position', typed_parser_1.seq((a1, _in, a2) => [a1, a2], primaryExpr, token_1.expectIdentifier('in'), primaryExpr));
    const substringFunction = specialFunctionParser('substring', typed_parser_1.seq((a1, _from, a2, a3) => (a3 ? [a1, a2, a3] : [a1, a2]), primaryExpr, token_1.expectIdentifier('from'), primaryExpr, utils_1.optional(typed_parser_1.seq(typed_parser_1.$2, token_1.expectIdentifier('for'), primaryExpr))));
    const trimDirection = typed_parser_1.oneOf(token_1.expectIdentifier('leading'), token_1.expectIdentifier('trailing'), token_1.expectIdentifier('both'));
    function trimArgs(dir, chrs, str) {
        return [
            ast_1.Expression.createConstant(dir || 'both'),
            chrs || ast_1.Expression.createConstant(''),
            str,
        ];
    }
    // trim([leading | trailing | both] from string [, characters] )
    // trim([leading | trailing | both] characters from string)
    // trim([leading | trailing | both] string [, characters] )
    const trimFunction = specialFunctionParser('trim', typed_parser_1.seq((dir, [str, chrs]) => trimArgs(dir, chrs, str), utils_1.optional(trimDirection), typed_parser_1.oneOf(typed_parser_1.seq((_from, str, chars) => [str, chars], token_1.expectIdentifier('from'), primaryExpr, utils_1.optional(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol(','), primaryExpr))), typed_parser_1.oneOf(typed_parser_1.attempt(typed_parser_1.seq((chrs, _from, str) => [str, chrs], primaryExpr, token_1.expectIdentifier('from'), primaryExpr)), typed_parser_1.seq((str, chrs) => [str, chrs], primaryExpr, utils_1.optional(typed_parser_1.seq(typed_parser_1.$2, token_1.symbol(','), primaryExpr)))))));
    return typed_parser_1.oneOf(overlayFunction, positionFunction, substringFunction, trimFunction);
}
exports.specialFunctionCall = specialFunctionCall;
