"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typed_parser_1 = require("typed-parser");
const constants_1 = require("../constants");
// Token parsers
// whitespace and comments
exports._ = typed_parser_1.seq(typed_parser_1.$null, typed_parser_1.skip('\\s*'), typed_parser_1.many(typed_parser_1.seq(typed_parser_1.$null, typed_parser_1.expectString('--'), typed_parser_1.stringBeforeEndOr('\n'), typed_parser_1.skip('\\s*'))));
function symbol(s) {
    return typed_parser_1.seq(typed_parser_1.$null, typed_parser_1.expectString(s, 'symbol'), exports._);
}
exports.symbol = symbol;
// Like symbol but doesn't skip whitespace
function symbolKeepWS(s) {
    return typed_parser_1.expectString(s, 'symbol');
}
exports.symbolKeepWS = symbolKeepWS;
exports.matchIdentifier = typed_parser_1.match('[a-zA-Z_][a-zA-Z0-9_]*');
exports.quotedEscape = typed_parser_1.seq(typed_parser_1.$2, symbolKeepWS('\\'), typed_parser_1.oneOf(typed_parser_1.keyword('"', '"'), typed_parser_1.keyword('\\', '\\')));
exports.quotedInner = typed_parser_1.seq((s, tail) => s + tail, typed_parser_1.stringBefore('[\\"]'), typed_parser_1.oneOf(typed_parser_1.seq((e, t) => e + t, exports.quotedEscape, typed_parser_1.lazy(() => exports.quotedInner)), typed_parser_1.constant('')));
exports.quotedIdentifier = typed_parser_1.seq(typed_parser_1.$2, symbolKeepWS('"'), exports.quotedInner, symbol('"'));
exports.identifier = typed_parser_1.seq(typed_parser_1.$1, typed_parser_1.attempt(typed_parser_1.oneOf(typed_parser_1.map((identifier, toError) => constants_1.sqlReservedWords.includes(identifier.toUpperCase())
    ? toError(`Expected an identifier, got reserved word ${identifier}`)
    : identifier, exports.matchIdentifier), exports.quotedIdentifier)), exports._);
function expectIdentifier(ident) {
    return typed_parser_1.seq(_ => ident, typed_parser_1.attempt(typed_parser_1.map((match, toError) => match.toLowerCase() !== ident.toLowerCase()
        ? toError(`Expected ${ident}, got ${match}`)
        : ident, exports.matchIdentifier)), exports._);
}
exports.expectIdentifier = expectIdentifier;
exports.reservedWord = (word) => {
    if (!constants_1.sqlReservedWords.includes(word))
        throw new Error(`INTERNAL ERROR: ${word} is not included in reservedWords`);
    return typed_parser_1.seq(typed_parser_1.$1, typed_parser_1.attempt(typed_parser_1.map((match, toError) => match.toUpperCase() !== word
        ? toError(`Expected ${word}, got ${match}`)
        : word, exports.matchIdentifier)), exports._);
};
exports.sepReserveds = (words) => typed_parser_1.attempt(typed_parser_1.seq(_ => words, ...words.split(/\s+/).map(word => typed_parser_1.seq(typed_parser_1.$null, exports.reservedWord(word), exports._)), exports._));
exports.anyOperator = typed_parser_1.seq(typed_parser_1.$1, typed_parser_1.match('[-+*/<>=~!@#%^&|`?]{1,63}'), exports._);
exports.operator = (op) => typed_parser_1.attempt(typed_parser_1.map((match, toError) => match != op ? toError(`Operator ${op} expected`) : match, exports.anyOperator));
exports.anyOperatorExcept = (exclude) => typed_parser_1.attempt(typed_parser_1.map((match, toError) => exclude.includes(match)
    ? toError(`Operator other than ${exclude.join(' ')} expected`)
    : match, exports.anyOperator));
