"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Either = require("fp-ts/lib/Either");
const sql = require("./sql");
function schemaClient(pgClient) {
    function getTable(schemaName, tableName) {
        return () => __awaiter(this, void 0, void 0, function* () {
            const result = yield sql.tableColumns(pgClient, {
                schemaName: schemaName || 'public',
                tableName,
            });
            if (result.length === 0) {
                return Either.left(`No such table: ${fullTableName(schemaName, tableName)}`);
            }
            return Either.right({
                name: tableName,
                columns: result.map(col => ({
                    hidden: col.attnum < 0,
                    name: col.attname,
                    nullable: !col.attnotnull,
                    type: col.atttypid,
                })),
            });
        });
    }
    function getEnums() {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield sql.enums(pgClient)).map(row => ({
                oid: row.oid,
                name: row.typname,
                labels: row.labels,
            }));
        });
    }
    function getArrayTypes() {
        return __awaiter(this, void 0, void 0, function* () {
            return (yield sql.arrayTypes(pgClient)).map(row => ({
                oid: row.oid,
                elemType: row.typelem,
            }));
        });
    }
    return { getTable, getEnums, getArrayTypes };
}
exports.schemaClient = schemaClient;
function fullTableName(schemaName, tableName) {
    return (schemaName ? schemaName + '.' : '') + tableName;
}
